from yta_editor_nodes.processor.base import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import get_input_size
from typing import Union
from abc import abstractmethod


class _TransitionProcessor(_ProcessorGPUAndCPU):
    """
    *Abstract class*
    
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some transition that will be
    done by CPU or GPU (at least one of the options)*

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        The `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU to process the input.
        """
        transition_processor_cpu, transition_processor_gpu = self._instantiate_cpu_and_gpu_processors(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = transition_processor_cpu,
            processor_gpu = transition_processor_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.

        This method must be implemented by each class.
        """
        pass

    def process(
        self,
        first_input: Union['moderngl.Texture', 'np.ndarray'],
        second_input: Union['moderngl.Texture', 'np.ndarray'],
        progress: float,
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True,
        **kwargs
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the transition between the given `first_input`
        and `second_input` depending on the `progress`
        provided, with GPU or CPU according to the `do_use_gpu`
        value.
        """
        output_size = (
            get_input_size(first_input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            first_input = self._prepare_input(
                input = first_input,
                do_use_gpu = do_use_gpu
            ),
            second_input = self._prepare_input(
                input = second_input,
                do_use_gpu = do_use_gpu
            ),
            progress = progress,
            output_size = output_size,
            **kwargs
        )


from yta_editor_nodes.processor.base import _ProcessorGPUAndCPU
from typing import Union
from abc import abstractmethod


class _NodeComplex(_ProcessorGPUAndCPU):
    """
    *For internal use only*

    This class must be inherited by the specific
    implementation of some complex node that will be
    done by CPU or GPU (at least one of the options)

    A complex node, which is a node made by other nodes,
    that is capable of processing inputs and obtain a
    single output by using the GPU or the CPU.

    This type of node is for complex modifications.
    """

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        node_complex_cpu, node_complex_gpu = self._instantiate_cpu_and_gpu_processors(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = node_complex_cpu,
            processor_gpu = node_complex_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        pass

    # @abstractmethod
    # def process(
    #     self,
    #     input: Union['np.ndarray', 'moderngl.Texture'],
    #     output_size: Union[tuple[int, int], None] = None,
    #     do_use_gpu: bool = True,
    #     **kwargs
    # ) -> Union['np.ndarray', 'moderngl.Texture']:
    #     """
    #     Process the provided 'input' with GPU or CPU 
    #     according to the internal flag.
    #     """
    #     pass
        
    def _process_common(
        self,
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        **kwargs
    ):
        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        # TODO: Transform the inputs if needed, based on
        # the type and the processor we got: numpy to
        # texture if GPU, texture to numpy if CPU

        # The **kwargs here will include inputs and other
        # parameters defined in the specific children
        return processor.process(
            output_size = output_size,
            **kwargs
        )


from yta_validation import PythonValidator
from typing import Union


VALID_EDITION_NODES = ['SerialTimelineNode', 'ParallelTimelineNode']
"""
The list of valid edition nodes as strings, to
compare easy to validate.
"""

def is_edition_node(
    node: Union['SerialTimelineNode', 'ParallelTimelineNode']
) -> bool:
    """
    Check if the provided `node` is an edition node, which
    has to be a SerialTimelineNode or a ParallelTimelineNode (by now).

    TODO: Update this in the future.
    """
    return PythonValidator.is_instance_of(node, VALID_EDITION_NODES)

def validate_is_edition_node(
    node: Union['SerialTimelineNode', 'ParallelTimelineNode']
) -> None:
    """
    Check if the provided `node` is an edition node, which
    has to be a `SerialTimelineNode` or a `ParallelTimelineNode` (by now), or
    raise an exception if not.
    """
    if not is_edition_node(node):
        raise Exception('The "node" provided is not an edition node.')


from yta_editor_nodes.timeline.parameter.evaluation_context import EvaluationContext
from yta_editor_nodes.timeline.parameter.abstract import ParameterSource


class ConstantValue(ParameterSource):
    """
    The parameter source in which the value is always
    the same. It is constant.
    """

    def __init__(
        self,
        value: any
    ):
        self.value: any = value
        """
        The constant value itself.
        """

    def evaluate(
        self,
        context: EvaluationContext
    ) -> any:
        # As it is always the same value, we don't care
        # about the context
        return self.value



from yta_editor_nodes.abstract import _ProcessorGPUAndCPU
from typing import Union
from abc import abstractmethod, ABC


class _NodeProcessor(_ProcessorGPUAndCPU, ABC):
    """
    *For internal use only*

    *Singleton class*

    This class must be inherited by the specific
    implementation of some effect that will be done by
    CPU or GPU (at least one of the options)

    A simple processor node that is capable of
    processing inputs and obtain a single output by
    using the GPU or the CPU.

    This type of node is for the effects and 
    transitions.
    """

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        The `opengl_context` will be passed to the GPU node
        if needed (and existing).
        """
        node_complex_cpu, node_complex_gpu = self._instantiate_cpu_and_gpu_processors(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = node_complex_cpu,
            processor_gpu = node_complex_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented in each class.
        """
        pass

    def _process_common(
        self,
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        **kwargs
    ):
        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        # The inputs must be transformed before reaching
        # this point

        # The **kwargs here will include inputs and other
        # parameters defined in the specific children

        return processor.process(
            output_size = output_size,
            **kwargs
        )


from yta_editor_nodes.processor.video.base import _VideoNodeProcessor
from typing import Union


class BreathingFrameVideoNodeProcessor(_VideoNodeProcessor):
    """
    The frame but as if it was breathing.
    """

    def __init__(
        self,
        opengl_context: 'moderngl.Context'
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_gpu.processor.video.breathing_frame import BreathingFrameVideoNodeProcessorGPU

        node_cpu = None
        node_gpu = BreathingFrameVideoNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True,
        t: float = 0.0,
        zoom: float = 0.05
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return super().process(
            input = input,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            t = t,
            zoom = zoom
        )


"""
Module for the effects and nodes that are made by
putting different nodes together.
"""
from yta_editor_nodes.complex.display_over_at import DisplayOverAtNodeComplex


__all__ = [
    'DisplayOverAtNodeComplex'
]


"""
Nodes that modify inputs to obtain outputs but
depending on a 't' time moment to adjust it to the
time of the video in which the input (a frame of a
video) is being edited. A movement effect is not 
edited the same when we are at the begining of the
video effect than when we are at the end.
"""
from yta_editor_nodes.processor.video.breathing_frame import BreathingFrameVideoNodeProcessor
from yta_editor_nodes.processor.video.waving_frame import WavingFrameVideoNodeProcessor


__all__ = [
    'BreathingFrameVideoNodeProcessor',
    'WavingFrameVideoNodeProcessor'
]


from typing import Any


class EvaluationContext:
    """
    The circunstances and attributes we need to be
    able to calculate the value of a parameter in
    this specific context.
    """

    def __init__(
        self,
        t: float,
        frame_index: int,
        node_outputs: dict[str, Any],
        curves: dict[str, Any],
        backend: str,
    ):
        self.t = t
        self.frame_index = frame_index
        self.node_outputs = node_outputs
        self.curves = curves
        self.backend = backend



"""
The transitions module, with all the classes that
are able to process transitions by using CPU, GPU
or both.

All the classes here will have an instance of the
specific CPU and/or GPU class that is able to run
the code by using either CPU or GPU. The user can
choose between GPU and CPU and that option will be
considered (only if available).

Note for the developer:
A class must have, at least, one specific 
processor (GPU is prefered).

TODO: This module doesn't use 't' but 'progress'
so it is not a child of 'processor.video', maybe
we should move it to be 'processor.transitions'
instead of 'processor.video.transitions'... (?)
"""
from yta_editor_nodes.processor.video.transitions.alphapedia_mask import AlphaPediaMaskTransitionProcessor
from yta_editor_nodes.processor.video.transitions.bars_falling import BarsFallingTransitionProcessor
from yta_editor_nodes.processor.video.transitions.circle_closing import CircleClosingTransitionProcessor
from yta_editor_nodes.processor.video.transitions.circle_opening import CircleOpeningTransitionProcessor
from yta_editor_nodes.processor.video.transitions.crossfade import CrossfadeTransitionProcessor
from yta_editor_nodes.processor.video.transitions.distorted_crossfade import DistortedCrossfadeTransitionProcessor
from yta_editor_nodes.processor.video.transitions.slide import SlideTransitionProcessor


__all__ = [
    'AlphaPediaMaskTransitionProcessor',
    'BarsFallingTransitionProcessor',
    'CircleClosingTransitionProcessor',
    'CircleOpeningTransitionProcessor',
    'CrossfadeTransitionProcessor',
    'DistortedCrossfadeTransitionProcessor',
    'SlideTransitionProcessor'
]


from yta_editor_nodes.blender.base import _Blender
from typing import Union


class MixBlender(_Blender):
    """
    A blender that uses a float value to mix the
    result with the original input as much as that
    value determines.
    """

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_blenders(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method has been created to be used in both the
        '__init__' and the '__reinit__' methods.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.blender.mix import MixBlenderCPU
        from yta_editor_nodes_gpu.blender.mix import MixBlenderGPU

        node_cpu = MixBlenderCPU()
        node_gpu = MixBlenderGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu

    def blend(
        self,
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        mix_weight: float = 1.0
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        return self.process(
            base_input = base_input,
            overlay_input = overlay_input,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            mix_weight = mix_weight
        )


"""
This module cocntains the nodes that are capable of
building the scene by positioning the inputs in
different positions.
"""
from yta_editor_nodes.compositor.displacement_with_rotation import DisplacementWithRotationNodeCompositor


__all__ = [
    'DisplacementWithRotationNodeCompositor'
]


from yta_editor_nodes.timeline.parameter.evaluation_context import EvaluationContext
from abc import ABC, abstractmethod
from typing import Any


class ParameterSource(ABC):
    """
    The source that allows us to calculate the
    value of a parameter, based on its context,
    for a specific `t` time moment (wich is
    included in the context).
    """

    @abstractmethod
    def evaluate(
        self,
        context: EvaluationContext
    ) -> Any:
        """
        Evaluate the value for this parameter based on the
        given `context`.
        """
        pass



"""
Old way of creating and handling effects. Now it
is done with the new node classes, so this whole
module is @deprecated and must be removed soon.
"""
# from yta_video_opengl.nodes.video import BreathingFrameVideoNodeProcessor, WavingFramesVideoNodeProcessor
# from yta_video_opengl.context import OpenGLContext
# from yta_video_opengl.nodes.audio import ChorusAudioNodeProcessor, VolumeAudioNodeProcessor
# from yta_video_opengl.nodes.timed import TimedVideoNode, TimedAudioNode
# from yta_video_opengl.utils import texture_to_frame
# from yta_validation.parameter import ParameterValidator
# from yta_validation import PythonValidator
# from yta_programming.singleton import SingletonMeta
# from typing import Union
# from abc import abstractmethod

# import moderngl


# class _AudioEffects:
#     """
#     *For internal use only*

#     The audio effects that will be available
#     throught our internal _Effects class to
#     wrap and make available all the audio
#     effects we want to be available.
#     """

#     def __init__(
#         self,
#         effects: 'Effects'
#     ):
#         self._effects: Effects = effects
#         """
#         The parent instance that includes this
#         class instance as a property.
#         """

#     """
#     Here below we expose all the effects
#     we want the users to have available to
#     be used.
#     """
#     def chorus(
#         self,
#         sample_rate: int,
#         depth: int = 0,
#         frequency: float = 0.25,
#         start: Union[int, float, 'Fraction'] = 0,
#         end: Union[int, float, 'Fraction', None] = None
#     ) -> TimedAudioNode:
#         return _create_timed_node(
#             ChorusAudioNodeProcessor(
#                 sample_rate = sample_rate,
#                 depth = depth,
#                 frequency = frequency
#             ),
#             start = start,
#             end = end
#         )

#     def volume(
#         self,
#         factor: callable,
#         start: Union[int, float, 'Fraction'] = 0,
#         end: Union[int, float, 'Fraction', None] = None
#     ) -> TimedAudioNode:
#         return _create_timed_node(
#             VolumeAudioNodeProcessor(
#                 factor_fn = factor
#             ),
#             start = start,
#             end = end
#         )
    
#     # TODO: Include definitive and tested audio
#     # effects here below

# class _VideoEffects:
#     """
#     *For internal use only*

#     The video effects that will be available
#     throught our internal _Effects class to
#     wrap and make available all the video
#     effects we want to be available.
#     """

#     def __init__(
#         self,
#         effects: 'Effects'
#     ):
#         self._effects: Effects = effects
#         """
#         The parent instance that includes this
#         class instance as a property.
#         """

#     """
#     Here below we expose all the effects
#     we want the users to have available to
#     be used.
#     """
#     def waving_node(
#         self,
#         # TODO: Maybe 'frame_size' (?)
#         size: tuple[int, int],
#         amplitude: float = 0.05,
#         frequency: float = 10.0,
#         speed: float = 2.0,
#         start: Union[int, float, 'Fraction'] = 0.0,
#         end: Union[int, float, 'Fraction', None] = None
#     ) -> TimedVideoNode:
#         """
#         TODO: Explain this effect better.

#         The 'start' and 'end' time moments are the
#         limits of the time range in which the effect
#         has to be applied to the frames inside that
#         time range. Providing start=0 and end=None
#         will make the effect to be applied to any
#         frame.
#         """
#         return _create_timed_node(
#             WavingFramesVideoNodeProcessor(
#                 # TODO: Maybe make this a parameter (?)
#                 do_use_gpu = True,
#                 amplitude = amplitude,
#                 frequency = frequency,
#                 speed = speed
#             ),
#             start = start,
#             end = end
#         )
    
#     def breathing_node(
#         self,
#         # TODO: Handle this
#         output_size: tuple[int, int],
#         zoom: float = 0.05,
#         start: Union[int, float, 'Fraction'] = 0.0,
#         end: Union[int, float, 'Fraction', None] = None
#     ) -> TimedVideoNode:
#         """
#         TODO: Explain this effect better.

#         The 'start' and 'end' time moments are the
#         limits of the time range in which the effect
#         has to be applied to the frames inside that
#         time range. Providing start=0 and end=None
#         will make the effect to be applied to any
#         frame.
#         """
#         return _create_timed_node(
#             BreathingFrameVideoNodeProcessor(
#                 # TODO: Maybe make this a parameter (?)
#                 do_use_gpu = True,
#                 zoom = zoom
#             ),
#             start = start,
#             end = end
#         )
    
#     # TODO: Include definitive and tested video
#     # effects here below

# class Effects(metaclass = SingletonMeta):
#     """
#     *Singleton class*

#     It is a singleton instance to have a
#     unique context for all the instances
#     that need it and instantiate this
#     class to obtain it. Here we group all
#     the nodes we have available for the
#     user.

#     This class is to simplify the access to
#     the effect nodes and also to have the
#     single context always available.

#     Even though we can have more effects,
#     this class is also the way we expose only
#     the ones we actually want to expose to 
#     the user.

#     The GPU will make the calculations in
#     parallel by itself, so we can handle a
#     single context to make the nodes share
#     textures and buffers.
#     """

#     def __init__(
#         self,
#         opengl_context: Union[moderngl.Context, None] = None
#     ):
#         """
#         If you provide an OpenGL context, it will be
#         used, or a new one will be created if the
#         'opengl_context' parameter is None.
#         """
#         ParameterValidator.validate_instance_of('opengl_context', opengl_context, moderngl.Context)

#         self.opengl_context = (
#             OpenGLContext().context
#             if opengl_context is None else
#             opengl_context
#         )
#         """
#         The opengl context that will be shared
#         by all the opengl nodes.
#         """
#         self.audio: _AudioEffects = _AudioEffects(self)
#         """
#         Shortcut to the audio effects that are
#         available.
#         """
#         self.video: _VideoEffects = _VideoEffects(self)
#         """
#         Shortcut to the video effects that are
#         available.
#         """

# def _create_timed_node(
#     node: Union['_AudioNodeProcessor', '_VideoNodeProcessor'],
#     start: Union[int, float, 'Fraction'],
#     end: Union[int, float, 'Fraction', None]
# ) -> Union[TimedVideoNode, TimedAudioNode]:
#     """
#     *For internal use only*

#     Create a TimedNode with the provided 'node' and
#     the 'start' and 'end' time moments.
#     """
#     # We could be other classes in the middle,
#     # because an OpenglNode inherits from
#     # other class
#     ParameterValidator.validate_mandatory_subclass_of('node', node, ['_AudioNodeProcessor', '_VideoNodeProcessor', '_OpenGLBase'])

#     timed_node_class = (
#         TimedAudioNode
#         if PythonValidator.is_subclass_of(node, '_AudioNodeProcessor') else
#         TimedVideoNode
#     )

#     # We have to create a node wrapper with the
#     # time range in which it has to be applied
#     # to all the frames
#     return timed_node_class(
#         node = node,
#         start = start,
#         end = end
#     )

# class _EffectStacked:
#     """
#     *For internal use only*

#     Class to wrap an effect that will be
#     stacked with an specific priority.

#     Priority is higher when lower value,
#     and lower when higher value.
#     """

#     @property
#     def is_audio_effect(
#         self
#     ) -> bool:
#         """
#         Flag to indicate if it is an audio effect
#         or not.
#         """
#         return self.effect.is_audio_node

#     @property
#     def is_video_effect(
#         self
#     ) -> bool:
#         """
#         Flag to indicate if it is a video effect
#         or not.
#         """
#         return self.effect.is_video_node

#     def __init__(
#         self,
#         effect: Union[TimedVideoNode, TimedAudioNode],
#         priority: int
#     ):
#         self.effect: Union[TimedVideoNode, TimedAudioNode] = effect
#         """
#         The effect to be applied.
#         """
#         self.priority: int = priority
#         """
#         The priority this stacked frame has versus
#         the other stacked effects.
#         """

# class _VideoEffectStacked(_EffectStacked):
#     """
#     *For internal use only*

#     Class to wrap a video effect that will
#     be stacked with a specific priority.

#     Priority is higher when lower value,
#     and lower when higher value.
#     """

#     def __init__(
#         self,
#         effect: TimedVideoNode,
#         priority: int
#     ):
#         ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedVideoNode)

#         super().__init__(
#             effect = effect,
#             priority = priority
#         )

# class _AudioEffectStacked(_EffectStacked):
#     """
#     *For internal use only*

#     Class to wrap an audio effect that will
#     be stacked with a specific priority.

#     Priority is higher when lower value,
#     and lower when higher value.
#     """

#     def __init__(
#         self,
#         effect: TimedAudioNode,
#         priority: int
#     ):
#         ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedAudioNode)
        
#         super().__init__(
#             effect = effect,
#             priority = priority
#         )

# # TODO: Move to another py file (?)
# class _EffectsStack:
#     """
#     *For internal use only*

#     *This class is to be inherited by the
#     specific video and audio classes.

#     Class to include a collection of effects
#     we want to apply in some entity, that 
#     will make easier applying them.

#     You can use this stack to keep the effects
#     you want to apply on a Media or on the
#     Timeline of your video editor.
#     """

#     @property
#     def copy(
#         self
#     ) -> '_EffectsStack':
#         """
#         Get a copy of this instance.
#         """
#         effects_stack = self.__class__()

#         for effect_sttacked in self._effects:
#             effects_stack.add_effect(
#                 effect = effect_sttacked.effect.copy,
#                 priority = effect_sttacked.priority
#             )

#         return effects_stack
    
#     @property
#     def number_of_effects(
#         self
#     ) -> int:
#         """
#         The number of effects that are stacked.
#         """
#         return len(self._effects)

#     @property
#     def effects(
#         self
#     ) -> list[_EffectStacked]:
#         """
#         The effects but ordered by 'priority' and
#         'start' time moment.
#         """
#         return sorted(
#             [
#                 effect
#                 for effect in self._effects
#             ],
#             key = lambda effect: (effect.priority, effect.effect.start)
#         )
    
#     @property
#     def lowest_priority(
#         self
#     ) -> int:
#         """
#         The priority of the effect with the lowest
#         one, or 0 if no video effects.
#         """
#         return min(
#             (
#                 effect.priority
#                 for effect in self.effects
#             ),
#             default = 0
#         )
    
#     def __init__(
#         self
#     ):
#         self._effects: list[Union[_AudioEffectStacked, _VideoEffectStacked]] = []
#         """
#         A list containing all the effects that
#         have been added to this stack, unordered.
#         """

#     def get_effects_at(
#         self,
#         t: Union[int, float, 'Fraction']
#     ) -> list[Union[TimedVideoNode, TimedAudioNode]]:
#         """
#         Get the effects, ordered by priority and the
#         'start' field, that must be applied within the
#         't' time moment provided because they are in
#         the [start, end) time range.
#         """
#         return [
#             effect.effect
#             for effect in self.effects
#             if effect.effect.is_within_time(t)
#         ]

#     def add_effect(
#         self,
#         effect: Union[TimedVideoNode, TimedAudioNode],
#         priority: Union[int, None] = None
#     ) -> 'EffectsStack':
#         """
#         Add the provided 'effect' to the stack with
#         the also given 'priority'.
#         """
#         ParameterValidator.validate_mandatory_instance_of('effect', effect, [TimedVideoNode, TimedAudioNode])
#         ParameterValidator.validate_positive_int('priority', priority, do_include_zero = True)

#         # TODO: What about the same effect added
#         # twice during the same time range? Can we
#         # allow it? It will be applied twice for
#         # specific 't' time moments but with 
#         # different attributes. is it ok (?)

#         # TODO: What if priority is already taken?
#         # Should we let some effects have the same
#         # priority (?)
#         priority = (
#             self.lowest_priority + 1
#             if priority is None else
#             priority
#         )

#         self._effects.append(_EffectStacked(
#             effect = effect,
#             priority = priority
#         ))

#         return self
    
#     # TODO: Create 'remove_effect'
    
#     @abstractmethod
#     def apply_effects_at(
#         self,
#         frame: Union['np.ndarray', moderngl.Texture],
#         t: Union[int, float, 'Fraction']
#     ) -> Union['np.ndarray', moderngl.Texture]:
#         """
#         Apply all the effects that must be applied
#         for the given `t` time moment to the also
#         provided `frame` (that must be the video or
#         audio frame of that time moment).
#         """
#         pass
    
# class VideoEffectsStack(_EffectsStack):
#     """
#     Class to include a collection of video effects
#     we want to apply in some entity, that will
#     make easier applying and organizing them.

#     You can use this stack to keep the video effects
#     you want to apply on a Media or on the Timeline
#     of your editor.
#     """

#     def add_effect(
#         self,
#         effect: TimedVideoNode,
#         priority: Union[int, None] = None
#     ) -> 'VideoEffectsStack':
#         """
#         Add the provided video 'effect' to the stack
#         with the also given 'priority'.
#         """
#         ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedVideoNode)

#         return super().add_effect(
#             effect = effect,
#             priority = priority
#         )
    
#     def apply_effects_at(
#         self,
#         frame: Union['np.ndarray', moderngl.Texture],
#         t: Union[int, float, 'Fraction']
#     ) -> Union['np.ndarray', moderngl.Texture]:
#         """
#         Apply all the effects that must be applied
#         for the given `t` time moment to the also
#         provided `frame` (that must be the video
#         frame of that time moment).
#         """
#         for effect in self.get_effects_at(t):
#             frame = effect.process(frame, t)

#         # TODO: Check when the frame comes as a
#         # Texture and when as a numpy array. I
#         # think when we apply an opengl node it
#         # is a texture, but we need to return it
#         # as a numpy, always
#         return (
#             texture_to_frame(frame)
#             if PythonValidator.is_instance_of(frame, moderngl.Texture) else
#             frame
#         )
    
# class AudioEffectsStack(_EffectsStack):
#     """
#     Class to include a collection of audio effects
#     we want to apply in some entity, that will
#     make easier applying and organizing them.

#     You can use this stack to keep the audio effects
#     you want to apply on a Media or on the Timeline
#     of your editor.
#     """

#     def add_effect(
#         self,
#         effect: TimedAudioNode,
#         priority: Union[int, None] = None
#     ) -> 'VideoEffectsStack':
#         """
#         Add the provided audio 'effect' to the stack
#         with the also given 'priority'.
#         """
#         ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedAudioNode)

#         return super().add_effect(
#             effect = effect,
#             priority = priority
#         )

#     def apply_effects_at(
#         self,
#         frame: Union['np.ndarray', moderngl.Texture],
#         t: Union[int, float, 'Fraction']
#     ) -> Union['np.ndarray', moderngl.Texture]:
#         """
#         Apply all the effects that must be applied
#         for the given `t` time moment to the also
#         provided `frame` (that must be the audio
#         frame of that time moment).
#         """
#         for effect in self.get_effects_at(t):
#             frame = effect.process(frame, t)

#         # Frame can only by a numpy array here
#         return frame


from yta_editor_nodes.compositor.base import _NodeCompositor
from yta_editor_nodes.utils import get_input_size
from typing import Union


class DisplacementWithRotationNodeCompositor(_NodeCompositor):
    """
    The frame, but moving and rotating over other frame.
    """

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.compositor.displacement_with_rotation import DisplacementWithRotationNodeCompositorCPU
        from yta_editor_nodes_gpu.compositor.displacement_with_rotation import DisplacementWithRotationNodeCompositorGPU

        node_cpu = DisplacementWithRotationNodeCompositorCPU()
        node_gpu = DisplacementWithRotationNodeCompositorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        position: tuple[int, int] = (1920 / 2 * 1.5, 1080 / 2 * 1.5),
        size: tuple[int, int] = (1920 / 2 * 1.5, 1080 / 2 * 1.5),
        rotation: int = 45
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        output_size = (
            get_input_size(base_input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            base_input = self._prepare_input(
                input = base_input,
                do_use_gpu = do_use_gpu
            ),
            overlay_input = self._prepare_input(
                input = overlay_input,
                do_use_gpu = do_use_gpu
            ),
            output_size = output_size,
            position = position,
            size = size,
            rotation = rotation
        )


from typing import Union


# TODO: I think this method is repeated somewhere else...
def get_input_size(
    input: Union['moderngl.Texture', 'np.ndarray']
) -> tuple[int, int]:
    """
    Get the size of the `input` provided, that can be
    a `moderngl.Texture` or a `numpy.ndarray`.

    Supported types:
    - moderngl.Texture
    - numpy.ndarray (H×W or H×W×C)
    """
    from yta_validation import PythonValidator

    if PythonValidator.is_instance_of(input, 'Texture'):
        return (input.width, input.height)
    elif PythonValidator.is_numpy_array(input):
        if input.ndim == 2:
            # Grayscale image: H × W
            height, width = input.shape

            return (width, height)

        if input.ndim == 3:
            # Color image: H × W × C
            height, width, _ = input.shape

            return (width, height)

        raise ValueError(
            f"Unsupported numpy array shape: {input.shape}. "
            "Expected 2D (H×W) or 3D (H×W×C)."
        )
    else:
        raise Exception('Unexpected input')


from yta_editor_nodes.processor.base import _NodeProcessor
from yta_editor_nodes.utils import get_input_size
from typing import Union
from abc import ABC


# TODO: ABC, do we need it (?)
class _VideoNodeProcessor(_NodeProcessor, ABC):
    """
    *For internal use only*
    
    *Abstract class*

    This class must be inherited by the specific
    implementation of some effect that will be done by
    CPU or GPU (at least one of the options).

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU, but for 
    video frames, including a `t` time moment parameter
    when processing.
    """

    def process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True,
        t: float = 0.0,
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        output_size = (
            get_input_size(input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            input = self._prepare_input(
                input = input,
                do_use_gpu = do_use_gpu
            ),
            output_size = output_size,
            t = t,
            **kwargs
        )


from yta_editor_nodes.processor.base import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import get_input_size
from typing import Union
from abc import abstractmethod


class _Blender(_ProcessorGPUAndCPU):
    """
    *For internal use only*

    This class must be inherited by the specific
    implementation of some blenders that will use
    CPU or GPU (at least one of the options).

    Class that is capable of mixing 2 different
    inputs by using the GPU or the CPU.
    """

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None] = None,
        **kwargs
    ):
        """
        The `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU to process the input.
        """
        blender_cpu, blender_gpu = self._instantiate_cpu_and_gpu_blenders(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = blender_cpu,
            processor_gpu = blender_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_blenders(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method must be implemented by each class.
        """
        pass

    def process(
        self,
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        # TODO: Maybe set it automatically based on the texture
        # size (?)
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        mix_weight: float = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        output_size = (
            get_input_size(base_input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        # TODO: Transform the inputs if needed, based on
        # the type and the processor we got: numpy to
        # texture if GPU, texture to numpy if CPU

        return processor.process(
            base_input = self._prepare_input(
                input = base_input,
                do_use_gpu = do_use_gpu
            ),
            overlay_input = self._prepare_input(
                input = overlay_input,
                do_use_gpu = do_use_gpu
            ),
            output_size = output_size,
            mix_weight = mix_weight,
            **kwargs
        )
    
    def blend_multiple_inputs(
        self,
        inputs: list[Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        mix_weights: Union[list[float], float] = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'inputs' with GPU or CPU
        according to the internal flag.
        """
        # TODO: What do we do with the 'output_size' (?)
        output_size = (
            get_input_size(inputs[0])
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        # TODO: Transform the inputs if needed, based on
        # the type and the processor we got: numpy to
        # texture if GPU, texture to numpy if CPU

        return processor.process_multiple_inputs(
            inputs = inputs,
            output_size = output_size,
            mix_weights = mix_weights,
            **kwargs
        )


from yta_editor_nodes.timeline import TimelineNode
from yta_editor_nodes.timeline.parameter.abstract import ParameterSource
from typing import Union


class SerialTimelineNode(TimelineNode):
    """
    A node that is executed from a specific input and
    generates a single output that is sent to the next
    node.

    Check the `.process` method definition of the node
    you add to see the parameters you need to execute
    it and pass them through the `parameters_sources`
    parameter.
    """

    def __init__(
        self,
        # TODO: Put the correct class
        # TODO: Maybe I need the 'node' in the TimelineNode class
        # TODO: Maybe I need a shortcut to 'is_gpu_available'...
        node: 'NodeProcessor',
        name: str,
        t_start: Union[int, float, 'Fraction'] = 0.0,
        t_end: Union[int, float, 'Fraction', None] = None,
        parameters_sources: dict[str, ParameterSource] = {}
    ):
        self.node: 'NodeProcessor' = node
        """
        The node to execute and to obtain the output from.
        """

        super().__init__(
            name = name,
            t_start = t_start,
            t_end = t_end,
            parameters_sources = parameters_sources
        )

    def _process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        t_timeline: float,
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output.
        """
        t_local = self._get_t(t = t_timeline)

        # Add 't' as dynamic because some NodeProcessor
        # don't actually expect it as parameter
        kwargs['t'] = t_local

        return self.node.process(
            input = input,
            **kwargs
        )


from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from typing import Union


class DistortedCrossfadeTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos,
    transforming the first one into the second one
    with a distortion in between.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        from yta_editor_nodes_gpu.processor.video.transitions.distorted_crossfade import DistortedCrossfadeTransitionNodeProcessorGPU

        node_cpu = None
        node_gpu = DistortedCrossfadeTransitionNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        first_input: Union['moderngl.Texture', 'np.ndarray'],
        second_input: Union['moderngl.Texture', 'np.ndarray'],
        progress: float,
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True,
        intensity: float = 1.0
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        return super().process(
            first_input = first_input,
            second_input = second_input,
            progress = progress,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            intensity = intensity
        )


"""
This module is old and must is @deprecated due to
the new TimelineGraph and node classes and must
be removed soon.
"""
from yta_editor_nodes.processor import _NodeProcessor, SelectionMaskNodeProcessor
from yta_editor_nodes.blender import _Blender, AlphaBlender, MixBlender
from yta_validation.parameter import ParameterValidator
from typing import Union

import numpy as np
import moderngl


# TODO: A transition class should have not selection
# mask and a opacity of 1.0 (or should be ignored
# also)
class Node:
    """
    A generic node that receives an input and returns
    an output when processed, having a selection mask
    to be able to apply the changes to only one part
    of the input, and an opacity to indicate the
    percentage of effect to apply in the input.
    """

    def __init__(
        self,
        node_processor: _NodeProcessor,
        # TODO: Wtf is AnimatedValue (?)
        parameters: dict[str, 'AnimatedValue'],
        selection_mask: Union[np.ndarray, moderngl.Texture, None] = None,
        mix_weight: float = 1.0
    ):
        self._node_processor: _NodeProcessor = node_processor
        """
        The node processor we need to apply the change
        to the input.
        """
        # TODO: Wtf is AnimatedValue (?)
        self._parameters: dict[str, 'AnimatedValue'] = parameters
        """
        The parameters we will use in the Node.
        """
        self._selection_mask: Union[np.ndarray, moderngl.Texture] = selection_mask
        """
        The mask to apply to the processed output in order
        to affect the original input only where and as
        much as the selection mask is indicating. Check the
        'selection_mask.py' module.
        """
        self._mix_weight: float = np.clip(mix_weight, 0, 1)
        """
        The mix_weight to be applied by the node process,
        which means how much the effect must be mixed
        with the original input, giving us the control
        on the result:

        - `mix_weight=1.0` The effect will be applied and
        mixed at a 100% with the original input.
        - `mix_weight=0.5` The effect will be applied but
        only at a 50%.

        This is useful to avoid modifying the code that
        applies the modification but only affect the
        result as much as we want.
        """

    def evaluate(
        self,
        input: np.ndarray,
        t: float,
        context
    ) -> np.ndarray:
        # if self._mix_weight.evaluate(t) == 0:
        #     return input
        if self._mix_weight == 0:
            return input

        kwargs = {
            # name: value.evaluate(t)
            name: value
            for name, value in self._parameters.items()
        }

        result = self._node_processor.process(
            input = input,
            **kwargs
        )

        if self._selection_mask is not None:
            # TODO: I don't know very well how to handle this
            # TODO: What is this context (?)
            result = context.selection_mask_processor.process(
                original_input = input,
                processed_input = result,
                selection_mask_input = self._selection_mask.evaluate(t)
            )

        # TODO: What is this context (?)
        result = context.mix_blender.process(
            base_input = input,
            overlay_input = result,
            # mix_weight = self._mix_weight.evaluate(t)
            mix_weight = self._mix_weight
        )

        return result

    def process(
        self,
        # TODO: What about the type (?)
        input: Union[np.ndarray, moderngl.Texture],
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union[np.ndarray, moderngl.Texture]:
        """
        *For internal use only*

        Common method to apply the processing to the
        given 'input' and set the output as the 'dtype'
        provided (or the one the 'input' had if None).

        This method will process the 'input' and then,
        if provided, apply it only to the selection 
        mask and also with the opacity set when
        instantiated.

        The result is clipped to the [0, 255] range.
        """
        if self._mix_weight == 0:
            return input.copy()
        
        dtype = (
            input.dtype
            if dtype is None else
            dtype
        )

        # Process the full input by the node processor
        result = self._node_processor.process(
            input = input.copy(),
            **kwargs
        )

        # Process the selection mask if needed
        if self._selection_mask is not None:
            result = SelectionMaskNodeProcessor(
                do_use_gpu = True
            ).process(
                original_input = input,
                processed_input = result,
                selection_mask_input = self._selection_mask
            )
            
        # Apply the mix_weight (percentage of effect)
        result = MixBlender(
            do_use_gpu = True
        ).process(
            base_input = input,
            overlay_input = result,
            mix_weight = self._mix_weight
        )

        # Return always as a float32 or Texture
        return result

        # return np.clip(
        #     a = result,
        #     a_min = 0,
        #     a_max = 255
        # ).astype(dtype)
    
# from abc import ABC, abstractmethod

# class NodeGraph(ABC):
#     """
#     *Abstract class*

#     The abstract class of the NodeGraph.
#     """

#     @abstractmethod
#     def evaluate(
#         self,
#         input,
#         t: float,
#         context
#     ):
#         pass

# class SerialGraph(NodeGraph):
#     """
#     A graph that will evaluate the node graphs
#     one after another. The input will be processed
#     by the first node and the output of this one
#     processed by the next node.
#     """

#     def __init__(
#         self,
#         nodes: list[Node]
#     ):
#         self.nodes: list[Node] = nodes
#         """
#         The list of nodes to use to evaluate the input
#         and process it.
#         """

#     def evaluate(
#         self,
#         input,
#         t: float, 
#         context
#     ):
#         """
#         Evaluate the `input` provided for the `t` time
#         moment also given, using the `context`.
#         """
#         for node in self.nodes:
#             input = node.evaluate(
#                 input = input,
#                 t = t,
#                 context = context
#             )

#         return input
    
# class ParallelGraph(NodeGraph):
#     """
#     A graph that will evaluate the input with the
#     different nodes in parallel, and then combine
#     all the results in a single output.

#     The first node in the `nodes` list will be the
#     first node to process the input.
#     """

#     def __init__(
#         self,
#         nodes: list[Node],
#         blender: _Blender
#     ):
#         self._nodes: list[Node] = nodes
#         """
#         *For internal use only*

#         The list of nodes to use to evaluate and process
#         the input.
#         """
#         self._blender: _Blender = blender
#         """
#         *For internal use only*

#         The blender to use to combine the outputs of the
#         different nodes.
#         """

#     def evaluate(
#         self,
#         input,
#         t: float,
#         context
#     ):
#         """
#         Evaluate the `input` provided for the `t` time
#         moment also given, using the `context`.
#         """
#         results = [
#             node.evaluate(input, t, context)
#             for node in self.nodes
#         ]

#         return self.blender.blend_multiple_inputs(results)
    
#     def process(
#         self,
#         # TODO: What about the type (?)
#         input: np.ndarray,
#         dtype: Union[np.dtype, None] = None,
#         **kwargs
#     # TODO: What about the output type (?)
#     ) -> np.ndarray:
#         """
#         Process the `input` with the nodes that belong
#         to this Parallel Node, with the provided `**kwargs`
#         (if some provided) and obtain a single mixed result.
#         """
#         results = [
#             node.process(
#                 input = input,
#                 dtype = dtype,
#                 **kwargs
#             )
#             for node in self._nodes
#         ]

#         return self._blender.blend_multiple_inputs(
#             inputs = results,
#             # TODO: Is this ok (?)
#             dtype = dtype
#         )



    
# TODO: Node types below. Maybe use a 'UsableNode' 
# or 'RunnableNode' class to identify them. These
# are the different ways to process an input and
# to obtain the result. These are the classes that
# can be added to process some input, and not the
# Nodes directly.
class SerialNode:
    """
    A single node that will receive an input, process
    it and return an output, that could perfectly used
    for the next serial node.
    """

    def __init__(
        self,
        node: Node
    ):
        ParameterValidator.validate_mandatory_instance_of('node', node, Node)

        self._node: Node = node
        """
        The single node that will process the input.
        """

    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the `input` with the single node that
        belongs to this Serial Node, with the provided
        `**kwargs` (if some provided) and obtain the
        single result.
        """
        return self._node.process(
            input = input,
            dtype = dtype,
            **kwargs
        )
    
class ParallelNodes:
    """
    A graph that will evaluate the input with the
    different nodes in parallel, and then combine
    all the results in a single output.

    The first node in the `nodes` list will be the
    first node to process the input.
    """

    def __init__(
        self,
        nodes: list[Node],
        blender: _Blender
    ):
        self._nodes: list[Node] = nodes
        """
        *For internal use only*

        The list of nodes to use to evaluate and process
        the input.
        """
        self._blender: _Blender = blender
        """
        *For internal use only*

        The blender to use to combine the outputs of the
        different nodes.
        """

    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the `input` with the nodes that belong
        to this Parallel Node, with the provided `**kwargs`
        (if some provided) and obtain a single mixed result.
        """
        results = [
            node.process(
                input = input,
                dtype = dtype,
                **kwargs
            )
            for node in self._nodes
        ]

        return self._blender.blend_multiple_inputs(
            inputs = results,
            # TODO: Is this ok (?)
            dtype = dtype
        )


"""
Our system is working based on layers (each
track is a layer), so we will not implement (at
least by now) any LayerNodes concept. I comment
it because of this above.
"""
# class LayerNode:
#     """
#     A node that will be used within a group of layer
#     nodes.
#     """

#     def __init__(
#         self,
#         node: Node,
#         blender: _Blender = AlphaBlender,
#         mix_weight: float = 1.0
#     ):
#         self._node: Node = node
#         """
#         The node that belongs to the layer and will be
#         used in the combination.
#         """
#         self._blender: _Blender = blender
#         """
#         The blender to mix the result with the input
#         provided.
#         """
#         self._mix_weight: float = mix_weight
#         """
#         The mix_weight to be applied when mixing it with
#         the input provided.
#         """

#     def process(
#         self,
#         # TODO: What about the type (?)
#         input: np.ndarray,
#         dtype: Union[np.dtype, None] = None,
#         **kwargs
#     # TODO: What about the output type (?)
#     ) -> np.ndarray:
#         """
#         *For internal use only*

#         Common method to apply the processing to the
#         given 'input' and set the output as the 'dtype'
#         provided (or the one the 'input' had if None).

#         This method will process the 'input' and then,
#         if provided, apply it only to the selection 
#         mask and also with the opacity set when
#         instantiated.

#         The result is clipped to the [0, 255] range.
#         """
#         layer_output = self._node.process(
#             input = input,
#             dtype = dtype,
#             **kwargs
#         )

#         # Mix the layer with the previous result
#         return self._blender.process(
#             base_input = input,
#             overlay_input = layer_output,
#             mix_weight = self._mix_weight,
#             dtype = dtype
#         )
    
# class LayerNodes:
#     """
#     A group of nodes that are ordered by priority
#     within layers, process the same input and mix
#     their result in a single output result.
#     """

#     def __init__(
#         self,
#         layer_nodes: list[LayerNode],
#         # TODO: This is an abstract base class but not the
#         # one we will receive
#         blender: _Blender
#     ):
#         ParameterValidator.validate_mandatory_list_of_these_instances('layer_nodes', layer_nodes, LayerNode)
#         # TODO: Check that validation is working
#         ParameterValidator.validate_mandatory_subclass_of('blender', blender, _Blender)

#         self._layer_nodes: list[LayerNode] = layer_nodes
#         """
#         The list of layer nodes that will process the same
#         input in parallel.
#         """
#         # TODO: This is an abstract base class but not the
#         # one we will receive
#         self._blender: _Blender = blender
#         """
#         The class capable of blending the results of
#         the different nodes that have been executed
#         in parallel over the same input.
#         """

#     def process(
#         self,
#         # TODO: What about the type (?)
#         input: np.ndarray,
#         dtype: Union[np.dtype, None] = None,
#         **kwargs
#     # TODO: What about the output type (?)
#     ) -> np.ndarray:
#         """
#         Process the `input` with the nodes that belong
#         to this Parallel Node, with the provided `**kwargs`
#         (if some provided) and obtain a single mixed result.
#         """
#         base = input.copy()

#         for layer_node in self._layer_nodes:
#             layer_output = layer_node.process(
#                 input = result,
#                 dtype = dtype,
#                 **kwargs
#             )

#             # TODO: Node can use GPU or CPU

#             # Blend over accumulated result
#             result = self._blender.process(
#                 base_input = base,
#                 overlay_input = layer_output,
#                 # TODO: What about the 'opacity' here (?)
#                 #opacity = node.opacity if hasattr(node, "opacity") else 1.0,
#                 #dtype = dtype
#             )

#         return result




#         for layer_node in self._layer_nodes:
#             output = layer_node.process(input)
#             # Blend it
#             output_blended = self._blender.blend([base, output])
#             # Adjust to this layer node opacity
#             output = (1 - layer_node.opacity) * base + layer_node.opacity * output_blended

#         # TODO: I'm doing this kind of operations in the 'blender.py'
#         return np.clip(
#             a = base,
#             a_min = 0,
#             a_max = 255
#         ).astype(dtype)